#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import pytest
from deltafi.exception import MissingMetadataException
from deltafi.input import EgressInput, TransformInput
from deltafi.storage import ContentService
from mockito import mock, unstub

from .helperutils import *


def test_egress_input():
    unstub()
    mock_content_service = mock(ContentService)
    event = make_event(mock_content_service)

    input = EgressInput(content=event.delta_file_messages[0].content_list,
                        metadata=event.delta_file_messages[0].metadata)
    assert input.has_content()


def test_egress_input_no_content():
    unstub()
    mock_content_service = mock(ContentService)
    event = make_event(mock_content_service)

    input = EgressInput(content=None,
                        metadata=event.delta_file_messages[0].metadata)
    assert input.has_content() is False


def test_transform_input():
    unstub()
    mock_content_service = mock(ContentService)
    event = make_event(mock_content_service)

    input = TransformInput(content=event.delta_file_messages[0].content_list,
                           metadata=event.delta_file_messages[0].metadata)

    assert input.content_named('CONTENT_NAME').name == "CONTENT_NAME"
    assert input.content_named('CONTENT 2').name == "CONTENT 2"
    assert input.content_named('CONTENT 3') is None

    assert input.get_metadata("plKey1") == "valueA"
    assert input.get_metadata_or_else("plkeyX", "not-found") == "not-found"
    with pytest.raises(MissingMetadataException):
        input.get_metadata("plkeyX")



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from deltafi.domain import Content, Context, Event
from deltafi.storage import ContentService, Segment
from mockito import mock, unstub

from .helperutils import *


def check_meta(meta):
    assert meta["key1"] == "value1"
    assert meta["key2"] == "value2"


def check_pl_meta(meta):
    assert meta["plKey1"] == "valueA"
    assert meta["plKey2"] == "valueB"


def make_content(name, content_service):
    return Content.from_dict(make_content_dict(name), content_service)


def test_content_json():
    content = make_content("CONTENT_NAME", make_context())
    content_json = content.json()
    assert content_json["name"] == "CONTENT_NAME"


def test_content_copy():
    segment = Segment(uuid="id1", offset=0, size=100, did="device1")
    content = Content(name="test", segments=[segment], media_type="text/plain", content_service=mock(ContentService))
    content.add_tag("tag")
    copied_content = content.copy()
    assert content == copied_content
    assert content is not copied_content
    assert content.segments is not copied_content.segments


def test_content_subcontent():
    segment = Segment(uuid="id1", offset=0, size=100, did="device1")
    content = Content(name="test", segments=[segment], media_type="text/plain", content_service=mock(ContentService))
    sub_content = content.subcontent(50, 25)
    assert sub_content.name == content.name
    assert sub_content.content_service == content.content_service
    assert len(sub_content.segments) == 1
    assert sub_content.segments[0].offset == 50
    assert sub_content.segments[0].size == 25


def test_content_get_size():
    segment = Segment(uuid="id1", offset=0, size=100, did="device1")
    content = Content(name="test", segments=[segment], media_type="text/plain", content_service=mock(ContentService))
    assert content.get_size() == 100


def test_content_get_set_media_type():
    segment = Segment(uuid="id1", offset=0, size=100, did="device1")
    content = Content(name="test", segments=[segment], media_type="text/plain", content_service=mock(ContentService))
    assert content.get_media_type() == "text/plain"
    content.set_media_type("application/json")
    assert content.get_media_type() == "application/json"


def test_content_prepend():
    content_service = mock(ContentService)

    segment1 = Segment(uuid="id1", offset=0, size=100, did="device1")
    content1 = Content(name="test1", segments=[segment1], media_type="text/plain", content_service=content_service)

    segment2 = Segment(uuid="id2", offset=0, size=200, did="device2")
    content2 = Content(name="test2", segments=[segment2], media_type="text/plain", content_service=content_service)

    content1.prepend(content2)
    assert len(content1.segments) == 2
    assert content1.segments[0].uuid == "id2"
    assert content1.segments[1].uuid == "id1"
    assert content1.get_size() == 300


def test_content_append():
    content_service = mock(ContentService)

    segment1 = Segment(uuid="id1", offset=0, size=100, did="device1")
    content1 = Content(name="test1", segments=[segment1], media_type="text/plain", content_service=content_service)

    segment2 = Segment(uuid="id2", offset=0, size=200, did="device2")
    content2 = Content(name="test2", segments=[segment2], media_type="text/plain", content_service=content_service)

    content1.append(content2)
    assert len(content1.segments) == 2
    assert content1.segments[0].uuid == "id1"
    assert content1.segments[1].uuid == "id2"
    assert content1.get_size() == 300


def test_segment_names():
    content_service = mock(ContentService)

    segment1 = Segment(uuid="id1", offset=0, size=100, did="123did")
    content1 = Content(name="test1", segments=[segment1], media_type="text/plain", content_service=content_service)

    segment2 = Segment(uuid="id2", offset=0, size=200, did="456did")
    content2 = Content(name="test2", segments=[segment2], media_type="text/plain", content_service=content_service)

    content1.append(content2)

    segment_names = content1.get_segment_names()
    assert len(segment_names) == 2
    assert "123/123did/id1" in segment_names
    assert segment_names['123/123did/id1'].size == 100
    assert "456/456did/id2" in segment_names
    assert segment_names['456/456did/id2'].size == 200


def test_context_json():
    unstub()
    mock_content_service = mock(ContentService)
    logger = None
    context = Context.create(make_context_dict(), mock_content_service, logger)

    assert context.did == TEST_DID
    assert context.delta_file_name == "FILENAME"
    assert context.flow_name == "FLOW_NAME"
    assert context.flow_id == "FLOW_ID"
    assert context.action_name == "ACTION_NAME"
    assert context.action_version == "1.0"
    assert context.hostname == "HOSTNAME"
    assert context.system_name == "SYSTEM_NAME"
    assert context.content_service == mock_content_service
    assert context.logger is None


def test_event():
    unstub()
    mock_content_service = mock(ContentService)
    logger = None
    event = Event.create({
        'deltaFileMessages': [make_delta_file_message_dict()],
        'actionContext': make_context_dict(),
        'actionParams': {}
    }, mock_content_service, logger)

    assert event.context.did == TEST_DID
    assert event.context.flow_name == "FLOW_NAME"
    assert event.context.action_name == "ACTION_NAME"
    assert event.context.content_service == mock_content_service
    assert event.context.logger is None
    assert len(event.params) == 0



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


import pytest
from deltafi.action import EgressAction
from deltafi.domain import Context
from deltafi.input import EgressInput
from deltafi.metric import Metric
from deltafi.result import EgressResult, ErrorResult, FilterResult
from deltafi.resultmessage import LogMessage, LogSeverity
from deltafi.test_kit.egress import EgressActionTest, EgressTestCase
from deltafi.test_kit.framework import IOContent
from pydantic import BaseModel


class SampleEgressAction(EgressAction):
    def __init__(self):
        super().__init__('Egress action for testing')

    def egress(self, context: Context, params: BaseModel, egress_input: EgressInput):
        data = egress_input.content.load_str()
        if 'error' in data:
            return ErrorResult(context, 'Failed to egress', 'error in content')
        elif 'filter' in data:
            return FilterResult(context, "Filtered")

        return (EgressResult(context)
                .add_metric(Metric(name="mKey", value=100))
                .log_info("my random message")
                .log_warning("my warning message"))


class SampleEgressActionTest(EgressActionTest):
    def __init__(self):
        super().__init__("test")
        self.action = SampleEgressAction()

    def egress_result(self, data: str, metric_key: str = "mKey", metric_val: int = 100):
        test_case = EgressTestCase(self.get_fields(data))
        test_case.expect_egress_result()
        test_case.add_metric(Metric(name=metric_key, value=metric_val))
        test_case.add_message(LogMessage.info('source', 'my .* message'))
        test_case.add_message(LogMessage.warning('source', 'my warning message'))
        if metric_key == "add_bogus":
            test_case.add_metric(Metric(name="bogus", value=1))
        self.egress(test_case)
        self.has_saved_content__size(0)

    def filter_result(self, data: str):
        test_case = EgressTestCase(self.get_fields(data))
        test_case.expect_filter_result("Filtered")
        self.egress(test_case)
        self.has_saved_content__size(0)

    def error_result(self, data: str):
        test_case = EgressTestCase(self.get_fields(data))
        test_case.expect_error_result("Failed to egress", "error in content")
        self.egress(test_case)
        self.has_saved_content__size(0)

    def get_fields(self, data: str):
        fields = {
            "data_dir": "none",
            "action": self.action,
            "inputs": [IOContent(file_name="input.txt", content_bytes=data)]
        }
        return fields


def test_egress_result():
    action_test = SampleEgressActionTest()
    action_test.egress_result("data")


def test_egress_result_bad_metric_value():
    action_test = SampleEgressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.egress_result("data", metric_val=0)
    assert "invalid metric value for mKey. E:0, A::100" in str(exc_info.value)


def test_egress_result_bad_metric_name():
    action_test = SampleEgressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.egress_result("data", metric_key="wrong")
    assert "invalid metric name. E:wrong, A::mKey" in str(exc_info.value)


def test_egress_result_wrong_metrics_count():
    action_test = SampleEgressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.egress_result("data", metric_key="add_bogus")
    assert "invalid metrics count. 2 != 1" in str(exc_info.value)


def test_not_an_egress_result():
    action_test = SampleEgressActionTest()
    with pytest.raises(ValueError) as exc_info:
        action_test.egress_result("error")
    assert "ErrorResult does not match EgressResult" in str(exc_info.value)


def test_filter_result():
    action_test = SampleEgressActionTest()
    action_test.filter_result("filter")


def test_not_a_filter_result():
    action_test = SampleEgressActionTest()
    with pytest.raises(ValueError) as exc_info:
        action_test.filter_result("error")
    assert "ErrorResult does not match FilterResult" in str(exc_info.value)


def test_error_result():
    action_test = SampleEgressActionTest()
    action_test.error_result("error")


def test_not_an_error_result():
    action_test = SampleEgressActionTest()
    with pytest.raises(ValueError) as exc_info:
        action_test.error_result("data")
    assert "EgressResult does not match ErrorResult" in str(exc_info.value)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


import pytest
from deltafi.action import TransformAction
from deltafi.domain import Context
from deltafi.input import TransformInput
from deltafi.result import TransformResult, TransformResults, ChildTransformResult, ErrorResult, FilterResult
from deltafi.resultmessage import LogMessage, LogSeverity
from deltafi.test_kit.framework import IOContent
from deltafi.test_kit.transform import TransformTestCase, TransformActionTest
from pydantic import BaseModel, Field


class SampleTransformParams(BaseModel):
    mode: str = Field(description="Choose the action operation mode")


class SampleTransformAction(TransformAction):
    def __init__(self):
        super().__init__('Transform action for testing')

    def param_class(self):
        return SampleTransformParams

    def transform(self, context: Context, params: SampleTransformParams, transform_input: TransformInput):
        # Creates orphaned content:
        TransformResult(context).save_string_content("123", "temp1.txt", "text/plain")
        if params.mode == "error":
            return ErrorResult(context, 'Error cause', 'error context')
        elif params.mode == "filter":
            return FilterResult(context, "Filter cause")
        else:
            data = f"{transform_input.first_content().load_str()}Output"
            if "meta" in params.mode:
                val = f"valueOut.{context.did}"
                return (TransformResult(context)
                        .add_metadata("keyOut", val)
                        .save_string_content(data, "output.txt", "text/plain"))
            elif "annotate" in params.mode:
                val = transform_input.get_metadata("keyIn")
                return (TransformResult(context)
                        .annotate("keyOut", val)
                        .delete_metadata_key("keyIn")
                        .save_string_content(data, "output.txt", "text/plain"))
            elif "many" in params.mode:
                transform_many_result = TransformResults(context)
                this_content = self.save_content("this", context)
                that_content = self.save_content("that", context)

                this_child = (ChildTransformResult(context, "this.many")
                              .add_content(this_content)
                              .add_metadata("key", "THIS"))
                that_child = (ChildTransformResult(context, "that.many")
                              .add_content(that_content)
                              .add_metadata("key", "THAT"))

                transform_many_result.add_result(this_child)
                transform_many_result.add_result(that_child)
                return transform_many_result
            elif "warning" in params.mode:
                return (TransformResult(context)
                        .save_string_content(data, "output.txt", "text/plain")
                        .log_warning("my warning message"))
            else:
                return (TransformResult(context)
                        .save_string_content(data, "output.txt", "text/plain"))

    def save_content(self, data: str, context):
        result = (TransformResult(context)
                  .save_string_content(data, f"{data}.txt", 'text/plain'))
        return result.content[0]


class SampleTransformActionTest(TransformActionTest):
    def __init__(self):
        super().__init__("test")
        self.action = SampleTransformAction()

    def transform_result(self,
                         param_mode: str = "transform",
                         output_name: str = "output.txt",
                         output_data: str = "InputOutput",
                         media_type: str = "text/plain",
                         del_meta_keys: list[str] = [],
                         meta_out: dict[str, str] = {},
                         annotations_out: dict[str, str] = {}):
        fields = {
            "data_dir": "none",
            "did": "abc123",
            "action": self.action,
            "file_name": "input.txt",
            "in_meta": {
                "keyIn": "valueIn"
            },
            "inputs": [
                IOContent(file_name="input.txt", content_bytes="Input")
            ],
            "parameters": {
                "mode": param_mode
            }
        }

        output = IOContent(
            file_name=output_name,
            content_bytes=output_data,
            content_type=media_type)

        test_case = TransformTestCase(fields)
        test_case.expect_transform_result()
        test_case.add_transform_result(
            content=[output],
            annotations=annotations_out,
            metadata=meta_out,
            delete_metadata_keys=del_meta_keys)
        if param_mode == 'warning':
            test_case.add_message(LogMessage.warning('source', 'my warning message'))
        self.transform(test_case)
        self.has_saved_content__size(2)

    def transform_many_result(self,
                              param_mode: str = "many",
                              num_outputs: int = 2):
        fields = {
            "data_dir": "none",
            "action": self.action,
            "file_name": "input.txt",
            "in_meta": {
                "keyIn": "valueIn"
            },
            "inputs": [
                IOContent(file_name="input.txt", content_bytes="Input")
            ],
            "parameters": {
                "mode": param_mode
            }
        }

        test_case = TransformTestCase(fields)
        test_case.expect_transform_results()
        test_case.add_transform_result(
            content=[IOContent(file_name="this.txt", content_bytes="this", content_type="text/plain")],
            annotations={},
            metadata={"key": "THIS"},
            delete_metadata_keys=[],
            name="this.many")

        if num_outputs == 2:
            test_case.add_transform_result(
                content=[IOContent(file_name="that.txt", content_bytes="that", content_type="text/plain")],
                annotations={},
                metadata={"key": "THAT"},
                delete_metadata_keys=[],
                name="that.many")

        self.transform(test_case)
        self.has_saved_content__size(3)

    def filter_result(self, filter_cause: str):
        fields = {
            "data_dir": "none",
            "did": "222",
            "action": self.action,
            "file_name": "input.txt",
            "inputs": [
                IOContent(file_name="input.txt", content_bytes="Input")
            ],
            "parameters": {
                "mode": "filter"
            }
        }

        test_case = TransformTestCase(fields)
        test_case.expect_filter_result(
            cause=filter_cause)
        self.transform(test_case)
        self.has_saved_content__size(1)

    def error_result(self, error_cause: str, error_context: str):
        fields = {
            "data_dir": "none",
            "action": self.action,
            "file_name": "input.txt",
            "inputs": [
                IOContent(file_name="input.txt", content_bytes="Input")
            ],
            "parameters": {
                "mode": "error"
            }
        }

        test_case = TransformTestCase(fields)
        test_case.expect_error_result(
            cause=error_cause,
            context=error_context)
        self.transform(test_case)
        self.has_saved_content__size(1)


def test_good_transform_many_result():
    action_test = SampleTransformActionTest()
    action_test.transform_many_result()


def test_not_a_transform_many_result():
    action_test = SampleTransformActionTest()
    with pytest.raises(ValueError) as exc_info:
        action_test.transform_many_result(param_mode="error")
    assert "ErrorResult does not match TransformResults" in str(exc_info.value)


def test_transform_many_result_missing_child():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_many_result(num_outputs=1)
    assert "invalid child count. 1 != 2" in str(exc_info.value)


def test_good_transform_result():
    action_test = SampleTransformActionTest()
    action_test.transform_result()


def test_good_transform_result_with_warning():
    action_test = SampleTransformActionTest()
    action_test.transform_result(param_mode='warning')


def test_not_a_transform_result():
    action_test = SampleTransformActionTest()
    with pytest.raises(ValueError) as exc_info:
        action_test.transform_result(param_mode="error")
    assert "ErrorResult does not match TransformResult" in str(exc_info.value)


def test_transform_result_wrong_output():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(output_data="Wrong")
    assert "Content[0]. Expected:\n<<Wrong>>\nBut was:\n<<InputOutput>>" in str(exc_info.value)


def test_transform_result_wrong_media_type():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(media_type="app/xml")
    assert "app/xml != text/plain" in str(exc_info.value)


def test_transform_result_wrong_name():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(output_name="file.xml")
    assert "file.xml != output.txt" in str(exc_info.value)


def test_good_transform_result_with_metadata():
    action_test = SampleTransformActionTest()
    action_test.transform_result(param_mode="meta", meta_out={"keyOut": "valueOut.abc123"})


def test_transform_result_missing_metadata():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(meta_out={"keyOut": "valueOut.abc123"})
    assert "keyOut not found" in str(exc_info.value)


def test_transform_result_wrong_meta_key():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(param_mode="meta", meta_out={"keyWrong": "valueOut.abc123"})
    assert "keyWrong not found" in str(exc_info.value)


def test_transform_result_too_much_metadata():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(param_mode="meta",
                                     meta_out={"keyOut": "valueOut.abc123", "this": "that"})
    assert "this not found" in str(exc_info.value)


def test_transform_result_wrong_meta_value():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(param_mode="meta", meta_out={"keyOut": "wrong"})
    assert "invalid value for key keyOut. E:wrong, A::valueOut.abc123" in str(exc_info.value)


def test_good_transform_result_with_annotations():
    action_test = SampleTransformActionTest()
    action_test.transform_result(param_mode="annotate",
                                 del_meta_keys=["keyIn"],
                                 annotations_out={"keyOut": "valueIn"})


def test_transform_result_ignore_dict_value():
    action_test = SampleTransformActionTest()
    action_test.transform_result(param_mode="annotate",
                                 del_meta_keys=["keyIn"],
                                 annotations_out={"keyOut": "%%IGNORE%%"})


def test_transform_result_missing_annotations():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(param_mode="annotate",
                                     del_meta_keys=["keyIn"],
                                     annotations_out={"wrongKey": "valueIn"})
    assert "wrongKey not found" in str(exc_info.value)


def test_transform_result_del_meta_keys_error():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.transform_result(param_mode="annotate",
                                     del_meta_keys=["key1", "key2"],
                                     annotations_out={"keyOut": "valueIn"})
    assert "key1 not found" in str(exc_info.value)


def test_good_filter_result():
    action_test = SampleTransformActionTest()
    action_test.filter_result("Filter cause")


def test_filter_result_wrong_cause():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.filter_result("wrong")
    assert "filter cause. Expected:\n<<wrong>>\nBut was:\n<<Filter cause>>" in str(exc_info.value)


def test_good_error_result():
    action_test = SampleTransformActionTest()
    action_test.error_result("Error cause", "error context")


def test_error_result_wrong_cause():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.error_result("wrong", "error context")
    assert "error cause. Expected:\n<<wrong>>\nBut was:\n<<Error cause>>" in str(exc_info.value)


def test_error_result_wrong_context():
    action_test = SampleTransformActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.error_result("Error cause", "wrong")
    assert "error context. Expected:\n<<wrong>>\nBut was:\n<<error context>>" in str(exc_info.value)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from unittest.mock import patch

import pytest
from deltafi.action import TransformAction
from deltafi.actiontype import ActionType
from deltafi.input import TransformInput
from deltafi.logger import get_logger
from deltafi.plugin import Plugin
from deltafi.result import TransformResult, EgressResult, ErrorResult
from deltafi.storage import ContentService
from mockito import when, mock, unstub
from pydantic import BaseModel, Field

from .helperutils import *


class SampleTransformParameters(BaseModel):
    thing: str = Field(description="An action parameter")


class SampleTransformAction(TransformAction):
    def __init__(self):
        super().__init__('A sample transform action')

    def param_class(self):
        return SampleTransformParameters

    def transform(self, context: Context, params: SampleTransformParameters, transform_input: TransformInput):
        # Creates two orphaned contents:
        TransformResult(context).save_string_content("123", "temp1.txt", "text/plain")
        TransformResult(context).save_string_content("abcdefg", "temp2.txt", "text/plain")

        return TransformResult(context) \
            .add_metadata('transformKey', 'transformValue') \
            .annotate('transformAnnotate', 'transformAnnotateValue') \
            .save_string_content("Abcde12345", "ten.txt", "text.plain", {"tag"})


class InvalidResult(TransformAction):
    def __init__(self):
        super().__init__('A sample transform action')

    def transform(self, context: Context, params: SampleTransformParameters, transform_input: TransformInput):
        return EgressResult(context)


class SampleErrorAction(TransformAction):
    def __init__(self):
        super().__init__('Create content but return error')

    def transform(self, context: Context, params: SampleTransformParameters, transform_input: TransformInput):
        # Creates orphaned content:
        TransformResult(context).save_string_content("123", "temp1.txt", "text/plain")

        return ErrorResult(context, 'Something bad happened', 'details')


@patch('time.time')
def test_action_returns_error(mock_time):
    unstub()
    mock_content_service = mock(ContentService)
    mock_time.return_value = 1754999744

    when(mock_content_service).put_str(...).thenReturn(make_segment('000'))
    when(mock_content_service).delete_all(...).thenReturn([])

    action = SampleErrorAction()
    event = make_event(mock_content_service)
    result = action.execute_action(event)
    assert type(result) == ErrorResult

    expected_response = {
        'annotations': {},
        'cause': 'Something bad happened',
        'context': 'details'
    }
    assert result.response() == expected_response

    plugin_to_response = Plugin.to_response(event, '12:00', '12:01', result)

    expected_plugin_to_response = {
        'actionName': 'ACTION_NAME',
        'did': '123did',
        'error': {
            'annotations': {},
            'cause': 'Something bad happened',
            'context': 'details'
        },
        'flowId': 'FLOW_ID',
        'flowName': 'FLOW_NAME',
        'messages': [
            {
                'severity': 'ERROR',
                'created': 1754999744,
                'source': 'ACTION_NAME',
                'message': "Something bad happened\ndetails"
            }
        ],
        'metrics': [],
        'start': '12:00',
        'stop': '12:01',
        'type': 'ERROR'
    }

    assert plugin_to_response == expected_plugin_to_response

    orphaned_content = Plugin.find_unused_content(event.context.saved_content, result)
    assert (len(orphaned_content)) == 1
    assert orphaned_content[0].uuid == "000"

    Plugin.orphaned_content_check(get_logger(), event.context, result, plugin_to_response)


def test_action_returns_transform():
    unstub()
    mock_content_service = mock(ContentService)

    when(mock_content_service).put_str(...) \
        .thenReturn(make_segment('111')) \
        .thenReturn(make_segment('222')) \
        .thenReturn(make_segment('333'))

    action = SampleTransformAction()
    assert action.action_type.value == ActionType.TRANSFORM.value
    event = make_event(mock_content_service)
    result = action.execute_action(event)
    assert type(result) == TransformResult

    expected_response = [{
        'did': '123did',
        'content': [
            {
                'mediaType': 'text.plain',
                'name': 'ten.txt',
                'segments': [
                    {
                        'did': '123did',
                        'offset': 0,
                        'size': 100,
                        'uuid': '333'
                    }
                ],
                'tags': ['tag']
            }
        ],
        'metadata': {
            'transformKey': 'transformValue'
        },
        'annotations': {
            'transformAnnotate': 'transformAnnotateValue'
        },
        'deleteMetadataKeys': []
    }]
    assert result.response() == expected_response

    orphaned_content = Plugin.find_unused_content(event.context.saved_content, result)
    assert (len(orphaned_content)) == 2
    assert orphaned_content[0].uuid == "111"
    assert orphaned_content[1].uuid == "222"


def test_invalid_result():
    unstub()
    mock_content_service = mock(ContentService)

    action = InvalidResult()
    with pytest.raises(ValueError):
        action.execute_action(make_event(mock_content_service))



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


import pytest
from deltafi.action import TimedIngressAction
from deltafi.domain import Context
from deltafi.result import IngressResult, IngressResultItem
from deltafi.resultmessage import LogMessage, LogSeverity
from deltafi.test_kit.framework import IOContent
from deltafi.test_kit.timed_ingress import TimedIngressTestCase, TimedIngressActionTest
from pydantic import BaseModel


class SampleTimedIngressAction(TimedIngressAction):
    def __init__(self):
        super().__init__('TTimed Ingress action for testing')

    def make_item(self, context: Context, name: str, index: int):
        ingress_item = IngressResultItem(context, name)
        ingress_item.add_metadata("index", str(index))
        ingress_item.annotate("a", "b")
        if name == "even":
            ingress_item.save_string_content("even-data", "even.0", "text/plain")
        else:
            ingress_item.save_string_content("odd-data-a", "odd.1", "text/plain")
            ingress_item.save_string_content("odd-data-b", "odd.2", "text/plain")
        return ingress_item

    def ingress(self, context: Context, params: BaseModel):
        index = 0
        if context.memo is not None:
            index = 1 + int(context.memo)

        ingress_result = IngressResult(context)

        ingress_result.add_item(self.make_item(context, "even", index))
        if index % 2 == 1:
            index += 1
            ingress_result.add_item(self.make_item(context, "odd", index))

        ingress_result.memo = str(index)
        ingress_result.execute_immediate = False
        ingress_result.status_message = "success"
        ingress_result.log_info("my info message")
        return ingress_result


class SampleTimedIngressActionTest(TimedIngressActionTest):
    def __init__(self):
        super().__init__("test")
        self.action = SampleTimedIngressAction()

    def ingress_result(self,
                       two_results: bool = False,
                       skip_odd: bool = False,
                       memo_override: str = "no",
                       message: str = "success"):
        memo = None
        expected_memo = "0"
        expected_meta_val = "0"
        if two_results:
            memo = "0"
            expected_memo = "2"
            expected_meta_val = "1"

        if memo_override != "no":
            expected_memo = memo_override

        fields = {
            "data_dir": "none",
            "did": "111",
            "action": self.action,
            "in_memo": memo
        }

        even_outputs = [IOContent(
            file_name="even.0",
            content_bytes="even-data",
            content_type="text/plain")]

        odd_outputs = [
            IOContent(
                file_name="odd.1",
                content_bytes="odd-data-a",
                content_type="text/plain"),
            IOContent(
                file_name="odd.2",
                content_bytes="odd-data-b",
                content_type="text/plain")
        ]

        test_case = TimedIngressTestCase(fields)
        test_case.expect_ingress_result(
            memo=expected_memo,
            status_message=message)
        test_case.add_ingress_result_item(
            name="even",
            content=even_outputs,
            metadata={"index": expected_meta_val},
            annotations={"a": "b"})
        test_case.add_message(LogMessage.info('source', 'my info message'))
        if two_results and not skip_odd:
            test_case.add_ingress_result_item(
                name="odd",
                content=odd_outputs,
                metadata={"index": "2"})

        self.ingress(test_case)
        if two_results:
            self.has_saved_content__size(3)
        else:
            self.has_saved_content__size(1)


def test_ingress_result():
    action_test = SampleTimedIngressActionTest()
    action_test.ingress_result()


def test_ingress_result_with_memo():
    action_test = SampleTimedIngressActionTest()
    action_test.ingress_result(two_results=True)


def test_ingress_result_wrong_memo():
    action_test = SampleTimedIngressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.ingress_result(memo_override="wrong")
    assert "invalid memo. E:wrong, A::0" in str(exc_info.value)


def test_ingress_result_item_count_mismatch():
    action_test = SampleTimedIngressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.ingress_result(two_results=True, skip_odd=True)
    assert "item count mismatch. 1 != 2" in str(exc_info.value)


def test_ingress_result_wrong_message():
    action_test = SampleTimedIngressActionTest()
    with pytest.raises(AssertionError) as exc_info:
        action_test.ingress_result(message="wrong")
    assert "invalid status_message. Expected:\n<<wrong>>\nBut was:\n<<success>>" in str(exc_info.value)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import time

import redis
from deltafi.actioneventqueue import ActionEventQueue
from mockito import when, mock, unstub

TEST_URL = "http://127.0.0.1:12345"
TEST_ACTION = "Action"
TEST_ITEM = "1234"


def test_action_event_queue_size():
    unstub()
    mock_pool = mock(redis.ConnectionPool)
    mock_conn = mock(redis.Redis)

    when(redis).ConnectionPool(...).thenReturn(mock_pool)
    when(redis).Redis(connection_pool=mock_pool).thenReturn(mock_conn)
    when(mock_conn).zcard(TEST_ACTION).thenReturn(1)

    service = ActionEventQueue(TEST_URL, 5, "password", None)
    assert service.size(TEST_ACTION) == 1


def test_action_event_queue_put():
    unstub()
    mock_pool = mock(redis.ConnectionPool)
    mock_conn = mock(redis.Redis)

    when(redis).ConnectionPool(...).thenReturn(mock_pool)
    when(redis).Redis(connection_pool=mock_pool).thenReturn(mock_conn)
    when(time).time().thenReturn(1.6)
    when(mock_conn).zadd(TEST_ACTION, {TEST_ITEM: 1600}, nx=True).thenReturn(1)

    service = ActionEventQueue(TEST_URL, 5, "password", None)
    assert service.put(TEST_ACTION, TEST_ITEM) == 1


def test_action_event_queue_take():
    unstub()
    mock_pool = mock(redis.ConnectionPool)
    mock_conn = mock(redis.Redis)

    result_tuple = (TEST_ACTION, TEST_ITEM, 1600)

    when(redis).ConnectionPool(...).thenReturn(mock_pool)
    when(redis).Redis(connection_pool=mock_pool).thenReturn(mock_conn)
    when(mock_conn).bzpopmin(TEST_ACTION, 0).thenReturn(result_tuple)

    service = ActionEventQueue(TEST_URL, 5, "password", "some-pod")
    assert service.take(TEST_ACTION) == TEST_ITEM



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import time
from enum import Enum
from typing import NamedTuple


class LogSeverity(Enum):
    TRACE = "TRACE"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    USER = "USER"


class LogMessage(NamedTuple):
    severity: LogSeverity
    created: int
    source: str
    message: str

    @classmethod
    def info(cls, source: str, message: str):
        return LogMessage(severity=LogSeverity.INFO, created=time.time(), source=source, message=message)

    @classmethod
    def warning(cls, source: str, message: str):
        return LogMessage(severity=LogSeverity.WARNING, created=time.time(), source=source,
                          message=message)

    @classmethod
    def error(cls, source: str, message: str):
        return LogMessage(severity=LogSeverity.ERROR, created=time.time(), source=source,
                          message=message)

    def json(self):
        return {'severity': self.severity.value,
                'created': self.created,
                'source': self.source,
                'message': self.message}
